#!/usr/bin/python2
# -*- coding utf-8 -*-

import trollius
from torrent import Torrent
from config import logger, PEER_REQUEST_LENGTH, EXTERNAL_PORT, UPNP, PEER_MAX_SIMILAR_CONN, PEER_SERVER_CONN_LIMIT, EXTERNAL_IP, PEER_FORCED_RECONNECT, DHT_INIT_NODES, DHT_EXTERNAL_PORT, DHT_SERVER
from web_server import make_web_app, web_server, EWaitPiece
from cache import Cache, cache_man
import math
import time
from http_client import HTTPClient
import upnp
from async_udp_announce import AsyncUDPAnnounce
from async_http_announce import AsyncHTTPAnnounce, get_peer_from_url
import ssl
from dht_server import DHTServer

LOOP_DEBUG = False

D_STATE_NONE = 0
D_STATE_DOWNLOAD = 1
D_STATE_CHECK = 2
D_STATE_WAIT = 3
D_STATE_CACHED = 4

if False:  # declare type def
    from peer_client import PeerClient


def cache_func(timeout):
    def get_f(f):
        d_cache = {'time': 0}

        def wrapper(*args):
            now = time.time()
            if now > d_cache['time']:
                d_cache['val'] = f(*args)
                d_cache['time'] = now + timeout
            return d_cache['val']

        return wrapper

    return get_f


class BadPiece:
    def __init__(self, engine, piece_index):
        # type: (LTorrentEngine, int) -> None
        self.engine = engine  # type: LTorrentEngine
        self.piece_index = piece_index
        self.parts_count = int(math.ceil(float(self.engine.cache.torrent.get_piece_size(piece_index)) / PEER_REQUEST_LENGTH))
        self.part_conditions = [D_STATE_NONE] * self.parts_count
        self.probes = tuple([] for i in range(self.parts_count))
        self.piece = self.engine.cache.read(piece_index)

    def error(self, message):
        logger.error('BadPiece: %s' % message)

    def debug(self, message):
        logger.debug('BadPiece: %s' % message)

    def info(self, message):
        logger.info('BadPiece: %s' % message)

    def get_piece_part(self, part_index):
        offset = part_index * PEER_REQUEST_LENGTH
        return self.piece[offset: offset + self.engine.cache.torrent.get_part_size(self.piece_index, part_index)]

    def is_candidate(self, peer):
        # type: (PeerClient) -> int
        for part in range(self.parts_count):
            if part not in self.engine.d_history[self.piece_index]:
                self.debug(part)
                self.debug(str(self.engine.d_history[self.piece_index].keys()))
                exit(0)
            if self.part_conditions[part] == D_STATE_NONE and self.engine.d_history[self.piece_index][part] != peer:
                for probe in self.probes[part]:
                    if probe[0] == peer:
                        break
                else:
                    if peer.bitfield[self.piece_index]:
                        self.part_conditions[part] = D_STATE_CHECK
                        self.probes[part].append([peer, None])
                        self.debug('candidate check %s %s' % (self.piece_index, part))
                        return part

    def update_probe(self, peer, data):
        # type: (PeerClient, str) -> None
        if self.piece_index != peer.download_piece_index:
            self.error('piece index do not match')
            return
        for probe in self.probes[peer.download_piece_part]:
            if probe[0] == peer:
                probe[1] = data
                break
        else:
            self.error('lost peer')
            return
        if data == self.get_piece_part(peer.download_piece_part):
            self.part_conditions[peer.download_piece_part] = D_STATE_CACHED
            self.debug('one shot approve %s %s' % (self.piece_index, peer.download_piece_part))
        else:
            for probe in self.probes[peer.download_piece_part]:
                if probe[0] != peer and probe[1] == data:
                    self.part_conditions[peer.download_piece_part] = D_STATE_CACHED
                    self.debug('approve %s %s, trys %s' % (self.piece_index, peer.download_piece_part, len(self.probes[peer.download_piece_part])))
                    break
            else:
                self.part_conditions[peer.download_piece_part] = D_STATE_NONE
                self.debug('bad probe %s %s' % (self.piece_index, peer.download_piece_part))
        for i in self.part_conditions:
            if i != D_STATE_CACHED:
                break
        else:
            for part in range(self.parts_count):
                if len(self.probes[part]) > 1:
                    if self.probes[part][-1][1] != self.get_piece_part(part):
                        bad_peer = self.engine.d_history[self.piece_index][part]  # type: PeerClient
                        if bad_peer.peername not in self.engine.d_black_peer_list:
                            bad_peer.self_close = True
                            self.engine.d_black_peer_list.add(bad_peer.peername)
                            self.info('add to black list %s' % str(bad_peer.peername))
                    self.engine.cache.write(self.piece_index, part * PEER_REQUEST_LENGTH, self.probes[part][-1][1])
            if self.engine.d_complete(self.piece_index):
                self.info('piece %s fixed' % self.piece_index)
                self.engine.d_bad_pieces.remove(self)
                return
            else:
                self.engine.d_bad_pieces.remove(self)
                del self.engine.d_piece_conditions[self.piece_index]
                del self.engine.d_piece_parts[self.piece_index]
                del self.engine.d_history[self.piece_index]
                self.engine.d_piece_state[self.piece_index] = D_STATE_NONE
                self.info('oops too many errors in pieces, try download again')
                return


class LTorrentEngine:
    def __init__(self):
        if getattr(self, 'loop', None) is None:
            self.loop = trollius.get_event_loop()  # type: trollius.BaseEventLoop
        if getattr(self, 'web_app', None) is None:
            self.web_app = make_web_app()
        if getattr(self, 'd_black_peer_list', None) is None:
            self.d_black_peer_list = set()
        self.web_server = web_server
        self.web_server.bind_engine(self)
        self.cache = None  # type: Cache
        self.f_pieces_info = None  # f_size, first_piece, last_piece, first_offset, last_size
        self.f_pos = 0
        self.f_size = 0
        self.f_skip_size = 0
        self.d_start_stop_pieces = [0, 0]
        self.d_piece_state = {}
        self.d_piece_parts = {}
        self.d_active_connections = []
        self.d_good_peers = set()
        self.d_connection_set = set()  # peers wait and connected
        self.d_piece_conditions = {}
        self.http_client = HTTPClient()
        self.slow_connections = set()
        self.status = D_STATE_NONE
        self.d_history = {}
        self.announce_task = None  # type: trollius.Task
        self.announcers = []  # url, timeout, planed announce
        self.d_bad_pieces = []
        self.dht_peers = set()
        if getattr(self, 'dht_server', None) is None:
            self.dht_server = None  # type: DHTServer
        self.dht_search_task = None  # type: trollius.Task

    def d_update_history(self, peer):
        # type: (PeerClient) -> None
        self.d_history[peer.download_piece_index][peer.download_piece_part] = peer

    def a_add_peer(self, peer_addr, c_from=None):
        self.d_start_peer(peer_addr, c_from=c_from)

    @cache_func(30)
    def get_active_connection_set(self):
        return set((i.peername[0], i.peername[1] if not i.is_server else i.ep_ext_port) for i in self.d_active_connections if i.status in (D_STATE_DOWNLOAD, D_STATE_CHECK) and (not i.is_server or i.ep_ext_port is not None))

    @cache_func(5)
    def get_avegare_request_wait(self):
        s_sum = s_count = 0
        for peer in self.d_active_connections:  # type: PeerClient
            if peer.status == D_STATE_DOWNLOAD and peer.stat['last_request_time'] > 0:
                s_sum += peer.stat['last_request_time']
                s_count += 1
        return 0 if s_count == 0 else s_sum / s_count

    def d_stop_download(self):
        del self.cache
        for i in self.d_active_connections:  # type: PeerClient
            i.transport.close()
        if self.announce_task is not None:
            self.announce_task.cancel()
        if self.dht_search_task is not None:
            self.dht_search_task.cancel()
        self.__init__()

    def debug(self, message):
        logger.debug('ENGINE %s' % message)

    def info(self, message):
        logger.info('ENGINE %s' % message)

    def exception(self, message):
        logger.exception('ENGINE %s' % message)

    @trollius.coroutine
    def d_reconnect_cood_peers(self):
        for peer_addr in self.d_good_peers - self.d_connection_set:
            self.a_add_peer(peer_addr, 'reconnect')

    @trollius.coroutine
    def _d_start_peer(self, peer_addr, c_from):
        from peer_client import PeerClient
        try:
            self.d_connection_set.add(peer_addr)
            yield self.loop.create_connection(lambda: PeerClient(self, False, c_from=c_from), peer_addr[0], peer_addr[1])
        except Exception, e:
            try:
                self.d_connection_set.remove(peer_addr)
            except KeyError:
                pass
            self.info('connect error %s' % str(e))

    def d_start_peer(self, peer_addr, c_from=None):
        if self.status == D_STATE_CACHED:
            return
        if peer_addr in self.d_black_peer_list:
            return
        if peer_addr[0] == EXTERNAL_IP:
            return
        if len(self.d_active_connections) > PEER_SERVER_CONN_LIMIT * 2:
            return
        if peer_addr in self.d_connection_set:
            return
        similar_count = len([i for i in self.d_active_connections if not i.is_server and i.peername == peer_addr])
        if PEER_MAX_SIMILAR_CONN <= similar_count:
            return
        self.loop.create_task(self._d_start_peer(peer_addr, c_from))

    @trollius.coroutine
    def re_announce(self):
        while self.status == D_STATE_DOWNLOAD or self.status == D_STATE_CACHED:
            if self.cache is None:
                return
            now = time.time()
            if self.status == D_STATE_DOWNLOAD:
                min_timeout = 600 if len(self.d_active_connections) > PEER_SERVER_CONN_LIMIT else 120
            else:
                min_timeout = 3600
            for a in self.announcers:
                if now > a[2]:
                    self.info('re announce %s' % a[0])
                    self._get_announce(a[0])
                    a[2] = now + min(a[1], min_timeout)
            yield self.d_reconnect_cood_peers()
            yield trollius.sleep(10)

    def start_dht_server(self):
        @trollius.coroutine
        def init_dht_nodes():
            yield trollius.From(self.dht_server.start_future)
            for addr in DHT_INIT_NODES:
                if not self.dht_server.bootstraped:
                    yield trollius.From(self.dht_server.bootstrap(addr))

        if self.dht_server is None:
            self.dht_server = DHTServer(self)
            self.loop.create_task(self.loop.create_datagram_endpoint(lambda: self.dht_server, local_addr=('0.0.0.0', DHT_EXTERNAL_PORT)))
            self.loop.create_task(init_dht_nodes())

    def start(self):
        self.loop.set_debug(LOOP_DEBUG)
        from peer_client import PeerClient
        if UPNP:
            try:
                upnp.del_port_forward(EXTERNAL_PORT)
            except:
                pass
            try:
                upnp.add_port_forward(EXTERNAL_PORT, EXTERNAL_PORT, 'LTorrentEngine')
                self.info('UPNP port forwarding enabled')
            except Exception, e:
                self.info('UPNP error %s' % e)
        else:
            logger.info('UPNP not found, for seeding, setup port forwarding on you router')
        self.loop.create_task(self.loop.create_server(lambda: PeerClient(self, True), '0.0.0.0', EXTERNAL_PORT))
        if DHT_SERVER:
            self.start_dht_server()
        logger.info('Start SeedingServer on port %s' % EXTERNAL_PORT)
        self.loop.run_forever()
        self.loop.close()

    def open_torrent(self, raw_data):
        if self.cache is not None:
            self.d_stop_download()
        torrent = Torrent(raw_data=raw_data)
        self.cache = cache_man.get_cache(torrent)
        return torrent.info_hash().encode('hex')

    def f_open(self, f_index):
        self.f_pos = 0
        if self.cache.torrent.multifile:
            self.f_size = self.cache.torrent.data['info']['files'][f_index]['length']
            self.f_skip_size = sum([self.cache.torrent.data['info']['files'][i]['length'] for i in range(f_index)])
        else:
            self.f_size = self.cache.torrent.full_size
            self.f_skip_size = 0

        first_piece = self.f_skip_size // self.cache.torrent.piece_length
        first_offset = self.f_skip_size % self.cache.torrent.piece_length
        a_size = self.f_size - (self.cache.torrent.piece_length - first_offset)
        last_piece = first_piece + int(math.ceil(float(a_size) / self.cache.torrent.piece_length))
        # last_size = a_size % self.torrent.piece_length
        self.f_pos = 0
        self.d_check(first_piece, last_piece)
        self.info('select file %s, piece range %s - %s' % (f_index, first_piece, last_piece))

    def f_read(self, size=1024 * 16):
        buf = ''
        tail = self.f_size - self.f_pos
        if size > tail:
            size = tail
        while size > 0:
            skip_size = self.f_skip_size + self.f_pos
            read_piece = skip_size // self.cache.torrent.piece_length
            if not self.cache.complete_pieces[read_piece]:
                raise EWaitPiece(read_piece)
            read_offset = skip_size % self.cache.torrent.piece_length
            read_max_size = self.cache.torrent.get_piece_size(read_piece) - read_offset
            read_size = read_max_size if size > read_max_size else size
            buf += self.cache.read(read_piece, read_offset, read_size)
            size -= read_size
            self.f_pos += read_size
        return buf

    def _read_piece(self):
        skip_size = self.f_skip_size + self.f_pos
        read_piece = skip_size // self.cache.torrent.piece_length
        return read_piece

    def f_seek(self, pos):
        self.f_pos = pos

    def d_check(self, first_piece, last_piece):
        if self.status == D_STATE_DOWNLOAD:
            return
        for i in range(first_piece, last_piece + 1):
            if not self.cache.complete_pieces[i]:
                self.d_start_stop_pieces = [i, last_piece]
                self.info('check, start download piece range %s - %s' % (i, last_piece))
                self.d_start()
                return
        else:
            self.info('check, file in cache')

    def d_start(self):
        self.d_piece_state = {}
        for i in range(self.d_start_stop_pieces[0], self.d_start_stop_pieces[1] + 1):
            self.d_piece_state[i] = D_STATE_CACHED if self.cache.complete_pieces[i] else D_STATE_NONE
        self.status = D_STATE_DOWNLOAD
        self.announce_task = self.loop.create_task(self.re_announce())
        self.get_announce()
        if self.dht_server is not None:
            self.dht_search_task = self.loop.create_task(self.dht_server.search_peers(self.cache.torrent.info_hash()))

    def d_next(self, peer_client):  # history collector
        # type: (PeerClient) -> (int, int, int) # state, download_piece, download_part
        result = self._d_next(peer_client)
        if result[0] == D_STATE_DOWNLOAD:
            if result[1] not in self.d_history:
                self.d_history[result[1]] = {}
        return result

    def _d_next(self, peer_client):  #
        # type: (PeerClient) -> (int, int, int) # state, download_piece, download_part
        if peer_client.download_piece_index != -1 \
                and self.d_piece_state[peer_client.download_piece_index] == D_STATE_DOWNLOAD \
                and peer_client.current_task != D_STATE_CHECK:
            peer_part_state = self.d_piece_parts[peer_client.download_piece_index]
            peer_part_state[peer_client.download_piece_part] = D_STATE_CACHED
            for i in range(len(peer_part_state)):
                if peer_part_state[i] != D_STATE_CACHED:
                    break
            else:
                if not self.d_complete(peer_client.download_piece_index):
                    self.info('wrog piece hash %s' % peer_client.download_piece_index)
                    self.d_piece_state[peer_client.download_piece_index] = D_STATE_CHECK
                    self.d_bad_pieces.append(BadPiece(self, peer_client.download_piece_index))
        for bad_piece in self.d_bad_pieces:  # type: BadPiece # check bad pieces
            c_part = bad_piece.is_candidate(peer_client)
            if c_part is not None:
                peer_client.bad_piece = bad_piece
                return D_STATE_CHECK, bad_piece.piece_index, c_part
        for peer in tuple(self.slow_connections):  # type: PeerClient  # check slow peers
            if peer.download_piece_index != -1 \
                    and self.d_piece_state[peer.download_piece_index] == D_STATE_DOWNLOAD \
                    and self.d_piece_parts[peer.download_piece_index][peer.download_piece_part] == D_STATE_DOWNLOAD \
                    and peer_client.bitfield[peer.download_piece_index] \
                    and peer_client.errors > 0 \
                    and len([i for i in self.d_active_connections if i.download_piece_index == peer.download_piece_index and i.download_piece_part == peer.download_piece_part]) < 2:
                self.info('speed up %s, %s' % (peer.download_piece_index, peer.download_piece_part))
                self.slow_connections.remove(peer)
                return D_STATE_DOWNLOAD, peer.download_piece_index, peer.download_piece_part
        read_piece = self._read_piece()
        for i in range(read_piece, self.d_start_stop_pieces[1] + 1) + range(self.d_start_stop_pieces[0], read_piece):
            if not self.cache.complete_pieces[i]:
                if peer_client.bitfield[i]:
                    if self.d_piece_state[i] == D_STATE_NONE:
                        self.d_piece_state[i] = D_STATE_DOWNLOAD
                        self.d_piece_parts[i] = [D_STATE_NONE] * int(math.ceil(float(self.cache.torrent.get_piece_size(i)) / PEER_REQUEST_LENGTH))
                    if self.d_piece_state[i] == D_STATE_DOWNLOAD:
                        for j_index, j in enumerate(self.d_piece_parts[i]):
                            if j == D_STATE_NONE:
                                self.d_piece_parts[i][j_index] = D_STATE_DOWNLOAD
                                return D_STATE_DOWNLOAD, i, j_index
        else:
            self.status = D_STATE_CACHED
            return D_STATE_CACHED, -1, -1

    def d_get_peer_by_piece_index(self, piece_index):
        for i in self.d_active_connections:  # type: PeerClient
            if i.download_piece_index == piece_index:
                return i

    def d_complete(self, piece_index):
        if self.cache.set_piece_complete(piece_index):
            self.d_piece_state[piece_index] = D_STATE_CACHED
            self.info('complete download piece %s' % piece_index)
            if piece_index in self.d_piece_conditions:
                self.d_piece_conditions[piece_index].notify_all()
                del self.d_piece_conditions[piece_index]
            del self.d_piece_parts[piece_index]
            if piece_index in self.d_history:
                del self.d_history[piece_index]
            return True
        return False

    def _get_announce(self, announce_url):
        try:
            peer_info = get_peer_from_url(announce_url)
        except ValueError:
            self.info('Bad announce url %s' % announce_url)
            return
        if announce_url[:3].lower() == 'udp':
            coro = self.loop.create_datagram_endpoint(lambda: AsyncUDPAnnounce(self, announce_url), remote_addr=(peer_info[0], peer_info[1]))
        else:
            coro = self.loop.create_connection(lambda: AsyncHTTPAnnounce(self, announce_url), peer_info[0], peer_info[1], None if not peer_info[2] else ssl.create_default_context())
        self.info('Get announce %s' % announce_url)
        self.loop.create_task(coro)

    def get_announce(self):
        announce_list = set()
        announce_list.add(self.cache.torrent.data['announce'])
        if 'announce-list' in self.cache.torrent.data:
            for i in self.cache.torrent.data['announce-list']:
                for j in i:
                    announce_list.add(j)
        for announce_url in announce_list:
            self._get_announce(announce_url)


def main():
    logger.info('Start LTorrentEngine')
    engine = LTorrentEngine()
    engine.start()


if __name__ == '__main__':
    main()
