from torrent import Torrent
from config import logger, MAX_CACHE_FILE_SIZE, WORK_DIR, CACHE_TORRENT_NAME, CACHE_FILE_INDEX_NAME, CACHE_PIECES_NAME, CACHE_BITFIELD_NAME, CACHE_STAT_UPDATE_TIMEOUT
import os.path
import os
import shelve
from file_bit_map import FileBitMap
import hashlib
import time


class PieceNotExist(Exception):
    pass


class CacheMan:
    def __init__(self):
        self.caches = {}

    def get_cache(self, torrent=None, info_hash=None):
        if torrent is not None:
            if torrent.info_hash() in self.caches:
                return self.caches[torrent.info_hash()]
            else:
                cache = Cache(torrent)
                self.caches[torrent.info_hash()] = cache
                return cache
        else:
            if info_hash in self.caches:
                return self.caches[info_hash]
            else:
                hex_info_hash = info_hash.encode('hex')
                for f_name in os.listdir(WORK_DIR):
                    if len(f_name) == 40 and f_name == hex_info_hash:
                        torrent = Torrent(fn=os.path.join(WORK_DIR, f_name, CACHE_TORRENT_NAME))
                        cache = Cache(torrent)
                        self.caches[torrent.info_hash()] = cache
                        return cache


class Cache:
    def __init__(self, torrent):
        # type: (Torrent) -> None
        if isinstance(torrent, str):
            self.cache_dir = os.path.join(WORK_DIR, torrent.encode('hex'))
            if not os.path.exists(self.cache_dir):
                raise Exception('cache dir not found, fix me!')
            torrent = Torrent(os.path.join(self.cache_dir, CACHE_TORRENT_NAME))
        # self.torrent = weakref.proxy(torrent)  # type: Torrent
        self.torrent = torrent  # type: Torrent
        self.cache_dir = os.path.join(WORK_DIR, self.torrent.info_hash().encode('hex'))
        if not os.path.exists(self.cache_dir):
            self.init_cache_dir()
        self.cache_file_index = shelve.open(os.path.join(self.cache_dir, CACHE_FILE_INDEX_NAME))
        if 'current_file_id' not in self.cache_file_index:
            self.cache_file_index['current_file_id'] = 0
        self._cache_files = {}
        self.complete_pieces = FileBitMap(os.path.join(self.cache_dir, CACHE_BITFIELD_NAME), self.torrent.piece_count)
        self.update_subscribers = set()
        self.stat = {
            'receive_bytes': self.cache_file_index.get('receive_bytes', 0),
            'receive_bytes_old': self.cache_file_index.get('receive_bytes', 0),
            'send_bytes': self.cache_file_index.get('send_bytes', 0),
            'send_bytes_old': self.cache_file_index.get('send_bytes', 0),
            'update_timer': time.time(),
            'receive_speed': 0,
            'send_speed': 0,
        }

    def left(self):
        return sum([self.torrent.get_piece_size(i) for i, c in enumerate(self.complete_pieces) if not c])

    def update_stat(self):
        now = time.time()
        if now - self.stat['update_timer'] < CACHE_STAT_UPDATE_TIMEOUT:
            return
        time_delta = now - self.stat['update_timer']
        self.stat['receive_speed'] = (self.stat['receive_bytes'] - self.stat['receive_bytes_old']) / time_delta
        self.stat['send_speed'] = (self.stat['send_bytes'] - self.stat['send_bytes_old']) / time_delta
        self.stat['receive_bytes_old'] = self.stat['receive_bytes']
        self.stat['send_bytes_old'] = self.stat['send_bytes']
        self.cache_file_index['receive_bytes'] = self.stat['receive_bytes']
        self.cache_file_index['send_bytes'] = self.stat['send_bytes']
        self.stat['update_timer'] = now

    def get_cache_file(self, file_id):
        if file_id not in self._cache_files:
            piece_fn = os.path.join(self.cache_dir, CACHE_PIECES_NAME % file_id)
            self._cache_files[file_id] = open(piece_fn, 'rb+') if os.path.exists(piece_fn) else open(piece_fn, 'wb+')
        return self._cache_files[file_id]

    def init_cache_dir(self):
        os.mkdir(self.cache_dir)
        with open(os.path.join(self.cache_dir, CACHE_TORRENT_NAME), 'wb') as f:
            f.write(self.torrent.raw_data)

    def get_piece_locate(self, piece_index):
        # type: (int) -> (int, int) # cache_file_id, file_offset
        piece_size = self.torrent.get_piece_size(piece_index)
        if str(piece_index) not in self.cache_file_index:
            f = self.get_cache_file(self.cache_file_index[str('current_file_id')])  # type: file
            file_offset = os.fstat(f.fileno()).st_size
            if file_offset + piece_size > MAX_CACHE_FILE_SIZE:
                self.cache_file_index['current_file_id'] += 1
                f = self.get_cache_file(self.cache_file_index[str('current_file_id')])  # type: file
                file_offset = os.fstat(f.fileno()).st_size
            f.truncate(file_offset + piece_size)
            self.cache_file_index[str(piece_index)] = (self.cache_file_index['current_file_id'], file_offset)
            self.cache_file_index.sync()
        return self.cache_file_index[str(piece_index)]

    def write(self, piece_index, offset, data):
        cache_file_id, file_offset = self.get_piece_locate(piece_index)
        f = self.get_cache_file(cache_file_id)
        f.seek(file_offset + offset)
        f.write(data)
        self.stat['receive_bytes'] += len(data)
        self.update_stat()

    def set_piece_complete(self, piece_index):
        data = self.read(piece_index)
        disk_hash = hashlib.sha1(data).digest()
        original_hash = self.torrent.get_piece_hash(piece_index)
        result = disk_hash == original_hash
        self.complete_pieces[piece_index] = result
        if result:
            for i in self.update_subscribers:  # type: PeerServer
                try:
                    i.send_have(piece_index)
                except:
                    logger.exception('CACHE Send have error')
        return result

    def read(self, piece_index, offset=0, size=None, update_stat=False):
        cache_file_id, file_offset = self.get_piece_locate(piece_index)
        f = self.get_cache_file(cache_file_id)
        f.seek(file_offset + offset)
        if size is None:
            size = self.torrent.get_piece_size(piece_index) - offset
        if update_stat:
            self.stat['send_bytes'] += size
            self.update_stat()
        return f.read(size)

    def check(self):
        for piece_index, complete in enumerate(self.complete_pieces):
            if complete:
                data = self.read(piece_index)
                disk_hash = hashlib.sha1(data).digest()
                original_hash = self.torrent.get_piece_hash(piece_index)
                if disk_hash != original_hash:
                    logger.debug('CACHE Check error piece %s' % piece_index)
                    return False
        logger.debug('CACHE check OK')
        return True


cache_man = CacheMan()


def test1():
    t = Torrent('ubuntu-18.04.1-desktop-amd64.iso.torrent')
    c = Cache(t)
    piece_index = 0
    print c.set_piece_complete(piece_index)
    piece_size = c.torrent.get_piece_size(piece_index)
    data = open('ubuntu-18.04.1-desktop-amd64.iso').read(piece_size)
    c.write(piece_index, 0, data)
    print c.set_piece_complete(piece_index)


def test2():
    t = Torrent('The.Good.Doctor.S02E04.1080p.rus.LostFilm.TV.mkv.torrent')
    cache = cache_man.get_cache(torrent=t)
    for i in cache.complete_pieces:
        print cache.complete_pieces[i]


if __name__ == '__main__':
    test2()
