import trollius
import random
import struct
from config import logger, DHT_ADDR, DHT_EXTERNAL_PORT, DHT_K_LIMIT, DHT_PEER_LIMIT, DHT_PEER_LIFE_TIME, DHT_PEER_PING_TIME, DHT_REQUEST_TIMEOUT, DHT_SEARCH_THREADS, DHT_SEARCH_QUERIES_LIMIT, DHT_INIT_NODES
import bencode
import os
import time
from collections import deque
import heapq
import socket

if False:
    from LTorrentEngine import LTorrentEngine


def int_to_bytes(val, b_len=160 / 8):
    data = []
    for i in range(b_len):
        data.insert(0, chr(val & 0xff))
        val >>= 8
    return bytes(''.join(data))


def bytes_to_int(b):
    val = 0
    for i in b:
        val <<= 8
        val += ord(i)
    return val


def xor_distance(a, b):
    return a ^ b


def unpack_nodes(data):
    for i in range(len(data) // 26):
        bin_node_id, ip, port = struct.unpack('!20s4sH', data[i * 26:i * 26 + 26])
        addr = socket.inet_ntoa(ip), port
        yield addr, bin_node_id


def pack_addr(addr):
    return struct.pack('!4sH', socket.inet_aton(addr[0]), addr[1])


def heapq_pop_ex(num, h):
    l = []
    try:
        for _ in range(num):
            l.append(heapq.heappop(h))
    except IndexError:
        pass
    return l


class Node:
    def __init__(self, n_id, peer_addr):
        # type: (long, tuple) -> None
        self.n_id = n_id
        self.peer_addr = peer_addr
        self.last_ping = time.time()


class KBucket:
    def __init__(self, id_range, k_limit):
        # type: (tuple, int) -> None
        self.id_range = id_range
        self.k_limit = k_limit
        self.nodes = {}
        self.spare_nodes = deque([], self.k_limit)

    def add(self, node, spare=False):
        # type: (Node, bool) -> bool
        if len(self.nodes) >= self.k_limit:
            if not spare:
                for i in self.spare_nodes:  # type: Node
                    if i.n_id == node.n_id:
                        break
                else:
                    self.spare_nodes.appendleft(node)
            return False
        self.nodes[node.n_id] = node
        return True

    def clean(self):
        now = time.time()
        for node in self.nodes.values():  # type: Node
            if now - node.last_ping > DHT_PEER_LIFE_TIME:
                del self.nodes[node.n_id]
                try:
                    spare_node = self.spare_nodes.popleft()  # type: Node
                    self.nodes[spare_node.n_id] = spare_node
                except IndexError:
                    pass

    def split(self):
        midpoint = sum(self.id_range) / 2
        clone = KBucket((midpoint + 1, self.id_range[1]), self.k_limit)
        self.id_range = self.id_range[0], midpoint
        self.spare_nodes = deque([], self.k_limit)
        for node in self.nodes.values():  # type: Node
            if not self.in_range(self.nodes[node.n_id]):
                clone.add(self.nodes[node.n_id])
                del self.nodes[node.n_id]
        return clone

    def in_range(self, node):
        # type: (Node) -> bool
        return self.id_range[0] <= node.n_id <= self.id_range[1]


class RouteTable:
    def __init__(self, k_limit=DHT_K_LIMIT, nodes_limit=DHT_PEER_LIMIT):
        self.nodes_limit = nodes_limit
        self.buckets = [KBucket((0, 2 ** 160), k_limit)]
        self.size = 0
        self.k_limit = k_limit

    def add(self, node):
        # type: (Node) -> None
        if self.size < self.nodes_limit:
            for b in self.buckets:  # type: KBucket
                if b.in_range(node):
                    if node.n_id not in b.nodes:
                        right_bucket = b
                        if len(b.nodes) >= self.k_limit:
                            b_clone = b.split()
                            self.buckets.insert(self.buckets.index(b) + 1, b_clone)
                            if b_clone.in_range(node):
                                right_bucket = b_clone
                        right_bucket.add(node)
                        self.update_size()
                        break
                    break

    def update_size(self):
        self.size = 0
        for b in self.buckets:  # type: KBucket
            self.size += len(b.nodes)

    def clean(self):
        for b in self.buckets:  # type: KBucket
            b.clean()
        self.update_size()

    def enumerate_nodes(self):
        nodes = []
        for b in self.buckets:  # type: KBucket
            nodes += b.nodes.values()
        return nodes

    def get_nodes(self, n_id_int, head=8):
        nodes = map(lambda n: (xor_distance(n_id_int, n.n_id), n), self.enumerate_nodes())
        heapq.heapify(nodes)
        if head:
            return heapq.nsmallest(head, nodes)
        return nodes


class DHTServer(trollius.DatagramProtocol):
    def __init__(self, engine):
        # type: (LTorrentEngine) -> None
        self.transport = None  # type: trollius.transports.DatagramTransport
        self.my_addr_int = DHT_ADDR
        self.my_addr_bin = int_to_bytes(self.my_addr_int)
        self.requests = {}
        self.loop = trollius.get_event_loop()  # type: trollius.BaseEventLoop
        self.start_future = trollius.Future()
        self.route_table = RouteTable()
        self.version = 'PP'
        self.ping_route_table_task = None
        self.engine = engine
        self.bootstraped = False

    def info(self, message):
        logger.info('DHT %s' % message)

    def debug(self, message):
        logger.debug('DHT %s' % message)

    def add_dht_node(self, addr):
        # type: (typle) -> trollius.Task
        @trollius.coroutine
        def _add_dht_node(addr):
            if not self.bootstraped:
                yield trollius.From(self.bootstrap(addr))
            else:
                yield trollius.From(self.ping(addr))

        return self.loop.create_task(_add_dht_node(addr))

    def connection_made(self, transport):
        self.transport = transport
        self.start_future.set_result(None)
        self.info('started, port %s, address %s' % (DHT_EXTERNAL_PORT, self.my_addr_int))
        if self.ping_route_table_task is None:
            self.ping_route_table_task = self.loop.create_task(self.ping_route_table())

    def request_timeout(self, transaction_id):
        # self.info('connection timeout')
        if transaction_id in self.requests:
            future = self.requests[transaction_id][0]  # type: trollius.Future
            future.set_result(None)
            del self.requests[transaction_id]

    def send_request(self, addr, method, params, timeout=DHT_REQUEST_TIMEOUT):
        transaction_id = os.urandom(8)
        packet = {
            't': transaction_id,
            'v': self.version,
            'y': 'q',
            'q': method,
            'a': params,
        }
        future = trollius.Future()
        self.requests[transaction_id] = (future, self.loop.call_later(timeout, self.request_timeout, transaction_id))
        self.transport.sendto(bencode.encode(packet), (addr[0], addr[1]))
        return future

    def send_responce(self, addr, transaction_id, params):
        packet = {
            't': transaction_id,
            'v': self.version,
            'y': 'r',
            'r': params,
            'ip': pack_addr(addr)
        }
        self.transport.sendto(bencode.encode(packet), (addr[0], addr[1]))

    def send_ping(self, addr, transaction_id):
        self.info('receive ping %s' % str(addr))
        self.send_responce(addr, transaction_id, {'id': self.my_addr_bin})

    def send_find_node(self, addr, transaction_id, bin_n_id):
        nodes = ''
        for distance, node in self.route_table.get_nodes(bytes_to_int(bin_n_id)):  # type: (long, Node)
            nodes += struct.pack('!20s4sH', int_to_bytes(node.n_id), socket.inet_aton(node.peer_addr[0]), node.peer_addr[1])
        self.send_responce(addr, transaction_id, {'id': self.my_addr_bin, 'nodes': nodes})

    def send_peers(self):
        pass

    def ping(self, addr, timeout=DHT_REQUEST_TIMEOUT):
        return self.send_request(addr, 'ping', {'id': self.my_addr_bin}, timeout=timeout)

    def find_node(self, addr, n_id_bin):
        return self.send_request(addr, 'find_node', {'id': self.my_addr_bin, 'target': n_id_bin})

    def get_peers(self, addr, info_hash_bin):
        return self.send_request(addr, 'get_peers', {'id': self.my_addr_bin, 'info_hash': info_hash_bin})

    @trollius.coroutine
    def bootstrap(self, addr):
        n_id_bin = os.urandom(20)
        data = yield trollius.From(self.ping(addr))
        if data is not None:
            data = yield trollius.From(self.get_nodes(n_id_bin))
            for addr, n_id_bin in data:
                self.add_dht_node(addr)
            self.bootstraped = True
            self.info('bootstraped')

    @trollius.coroutine
    def ping_route_table(self):
        while True:
            now = time.time()
            for node in self.route_table.enumerate_nodes():  # type: Node
                if now - node.last_ping > DHT_PEER_PING_TIME:
                    data = yield trollius.From(self.ping(node.peer_addr))
                    if data is not None:
                        node.last_ping = now
            self.route_table.clean()
            yield trollius.sleep(DHT_PEER_PING_TIME // 4)

    @trollius.coroutine
    def find_peers(self, info_hash_bin, addr_list):  # addr_list = map(lambda i: i[1].peer_addr, self.route_table.get_nodes(bytes_to_int(info_hash_bin)))
        req = map(lambda i: self.get_peers(i, info_hash_bin), addr_list)
        data = yield trollius.From(trollius.gather(*req))
        nodes = []
        peers = []
        for i in data:
            if i is not None:
                if 'nodes' in i[1]['r']:
                    nodes += unpack_nodes(i[1]['r']['nodes'])
                elif 'values' in i[1]['r']:
                    if len(i[1]['r']['values']) > 0:
                        logger.info('found values %s' % str(i[1]['r']['values']))
                        for j in i[1]['r']['values']:
                            ip, port = struct.unpack('!4sH', j)
                            peers.append((socket.inet_ntoa(ip), port))
                else:
                    self.info('find_peers unknown packet %s' % str(i[1]))
        raise trollius.Return((nodes, peers))

    @trollius.coroutine
    def get_nodes(self, n_id_bin, addr_list=None):  # list of (addr, node_id_bin)
        if not addr_list:
            addr_list = []
            t_id = bytes_to_int(n_id_bin)
            nodes = map(lambda n: (xor_distance(t_id, n.n_id), n), self.route_table.enumerate_nodes())
            heapq.heapify(nodes)
            for distance, node in heapq.nsmallest(50, nodes):  # type: (long, Node)
                addr_list.append(node.peer_addr)
        req = [self.ping(addr) for addr in addr_list]
        data = yield trollius.From(trollius.gather(*req))
        req = [self.find_node(i[0], n_id_bin) for i in data if i is not None]
        data = yield trollius.From(trollius.gather(*req))
        nodes = []
        for i in data:
            if i is not None:
                nodes += unpack_nodes(i[1]['r']['nodes'])
        raise trollius.Return(nodes)

    def datagram_received(self, data, addr):
        try:
            result_dict = bencode.decode(data)
        except bencode.BencodeDecodeError:
            self.info('decode error %s' % data)
            return
        # self.info('datagram_received %s' % str(result_dict))
        if 'y' not in result_dict:
            self.info('not found "y" in %s' % str(result_dict))
            return
        if result_dict['y'] == 'r':
            if 'r' in result_dict and 'id' in result_dict['r']:
                n_id = bytes_to_int(result_dict['r']['id'])
                self.route_table.add(Node(n_id, addr))
            if 't' in result_dict and result_dict['t'] in self.requests:
                self.requests[result_dict['t']][1].cancel()
                self.requests[result_dict['t']][0].set_result((addr, result_dict))
                del self.requests[result_dict['t']]
            else:
                if 't' in result_dict:
                    self.info('receive old packet %s' % str(result_dict))
                else:
                    self.info('receive unknown packet %s' % str(result_dict))
        elif result_dict['y'] == 'q':
            if result_dict['q'] == 'ping':
                self.send_ping(addr, result_dict['t'])
        else:
            self.info('receive unknown packet %s' % str(result_dict))

    @trollius.coroutine
    def search_peers(self, info_hash_bin):
        if not self.bootstraped:
            for bs_addr in DHT_INIT_NODES:
                if not self.bootstraped:
                    yield trollius.wait(self.add_dht_node(bs_addr))
        if not self.bootstraped:
            self.info('search unvaliable, no bootstrap nodes')
        yield trollius.sleep(2)  # wait announcer
        t_id = bytes_to_int(info_hash_bin)
        queried = set()
        addr_list = map(lambda i: i[1].peer_addr, self.route_table.get_nodes(bytes_to_int(info_hash_bin), head=DHT_SEARCH_THREADS))
        h = []
        peers = set()
        req = set()
        stat_timer = time.time()
        while True:
            for i in addr_list:
                req.add(self.get_peers(i, info_hash_bin))
            queried.update(addr_list)
            done, req = yield trollius.wait(req, return_when=trollius.FIRST_COMPLETED)  # type: set, set
            for future in done:  # type: trollius.Future
                result = future.result()
                if result is not None:
                    res_addr, res = result
                    if 'nodes' in res['r']:
                        for n_addr, n_bin_n_id in unpack_nodes(res['r']['nodes']):
                            if n_addr not in queried:
                                heapq.heappush(h, (xor_distance(t_id, bytes_to_int(n_bin_n_id)), n_addr))
                    elif 'values' in res['r']:
                        self.info('found values %s' % str(res['r']['values']))
                        for j in res['r']['values']:
                            ip, port = struct.unpack('!4sH', j)
                            peer = (socket.inet_ntoa(ip), port)
                            if peer not in peers:
                                self.engine.a_add_peer(peer, c_from='DHT')
                            peers.add(peer)
                    else:
                        self.info('find_peers unknown packet %s' % str(res['r']))
            if len(req) == 0 and len(h) == 0:
                break
            if len(queried) > DHT_SEARCH_QUERIES_LIMIT:
                break
            addr_list = []
            if len(req) < DHT_SEARCH_THREADS and len(h) > 0:
                for distance, h_addr in heapq_pop_ex(DHT_SEARCH_THREADS - len(req), h):
                    addr_list.append(h_addr)
            if time.time() - stat_timer > 5:
                min_dist = heapq.nsmallest(1, h)
                self.info('queried %s, heap %s, req %s, peers %s, min dist: %s' % (len(queried), len(h), len(req), len(peers), str(min_dist)))
                stat_timer = time.time()
        self.info('search done')


def test1():
    test_num = random.randint(0, 99999999999999999999999999999999999999)
    print test_num
    print int_to_bytes(test_num).encode('hex')
    try:
        print struct.pack('!L', test_num).encode('hex')
    except:
        pass
    print bytes_to_int(int_to_bytes(test_num))


def test2():
    from torrent import Torrent
    dht_server = DHTServer()

    @trollius.coroutine
    def worked_task():
        # t = Torrent(fn='tmp/[rutor.is]Daredevil.S03.2018.WEBRip.2160p.MediaClub.torrent')
        # t = Torrent(fn='tmp/[rutor.is]The_Good_Doctor_s02_LF.torrent')
        t_id = bytes_to_int(t.info_hash())
        addr = ('178.237.176.172', 49638)
        yield trollius.From(dht_server.start_future)
        data = yield trollius.From(dht_server.ping(addr))

        queried = set()
        addr_list = map(lambda i: i[1].peer_addr, dht_server.route_table.get_nodes(bytes_to_int(t.info_hash()), head=DHT_SEARCH_THREADS))
        h = []
        peers = set()
        req = set()
        stat_timer = time.time()
        while True:
            for i in addr_list:
                req.add(dht_server.get_peers(i, t.info_hash()))
            queried.update(addr_list)
            done, req = yield trollius.wait(req, return_when=trollius.FIRST_COMPLETED)  # type: set, set
            for future in done:  # type: trollius.Future
                result = future.result()
                if result is not None:
                    res_addr, res = result
                    if 'nodes' in res['r']:
                        for n_addr, n_bin_n_id in unpack_nodes(res['r']['nodes']):
                            if n_addr not in queried:
                                heapq.heappush(h, (xor_distance(t_id, bytes_to_int(n_bin_n_id)), n_addr))
                    elif 'values' in res['r']:
                        logger.info('found values %s' % str(res['r']['values']))
                        for j in res['r']['values']:
                            ip, port = struct.unpack('!4sH', j)
                            peers.add((socket.inet_ntoa(ip), port))
                    else:
                        logger.info('find_peers unknown packet %s' % str(res['r']))
            if len(req) == 0 and len(h) == 0:
                break
            if len(queried) > 5000:
                break
            addr_list = []
            if len(req) < DHT_SEARCH_THREADS and len(h) > 0:
                for distance, h_addr in heapq_pop_ex(DHT_SEARCH_THREADS - len(req), h):
                    addr_list.append(h_addr)
            if time.time() - stat_timer > 3:
                min_dist = heapq.nsmallest(1, h)
                logger.info('queried %s, heap %s, req %s, peers %s, min dist: %s' % (len(queried), len(h), len(req), len(peers), str(min_dist)))
                stat_timer = time.time()
        logger.info('Search done')

    loop = trollius.get_event_loop()  # type: trollius.BaseEventLoop
    loop.set_debug(True)
    loop.create_task(loop.create_datagram_endpoint(lambda: dht_server, local_addr=('0.0.0.0', DHT_EXTERNAL_PORT)))
    loop.create_task(worked_task())
    loop.run_forever()
    loop.close()


def test4():
    def print_rt(f_rt):
        # type: (RouteTable) -> None
        for b in f_rt.buckets:
            print 'limit %s<>%s' % (b.id_range[0], b.id_range[1])
            for n in b.nodes:  # type: Node
                print '  node %s' % n

    rt = RouteTable()
    for i in range(1024):
        r_id = random.randint(0, 2 ** 160)
        rt.add(Node(r_id, ('127.0.0.1', 8000)))
    # print_rt(rt)
    print_rt(rt)
    print rt.size


def test5():
    import timeit
    id1 = os.urandom(20)
    id2 = os.urandom(20)

    id1b = bytes_to_int(id1)
    id2b = bytes_to_int(id2)

    def get_d1():
        return ''.join([chr(ord(i1) ^ ord(i2)) for i1, i2 in zip(id1, id2)])

    def get_d2():
        return id1b ^ id2b

    print timeit.timeit(stmt=get_d1, number=100000)
    print timeit.timeit(stmt=get_d2, number=100000)


if __name__ == '__main__':
    test5()
