import tornado
import tornado.web
import tornado.escape
import tornado.gen
from tornado.locks import Condition
from tornado.iostream import StreamClosedError
import os.path
from config import MY_DIR, logger, WEB_SERVER_PORT, WEB_SERVER_SEND_BUF
from tornado.platform.asyncio import AsyncIOMainLoop
import json
from http_client import HTTPClient
import re
import time

AsyncIOMainLoop().install()

if False:  # declare type def
    from LTorrentEngine import LTorrentEngine

re_parce_range = re.compile('bytes=(\d*)-(\d*)')


class EWaitPiece(Exception):
    def __init__(self, wait_piece):
        self.wait_piece = wait_piece


class WebServer:
    def __init__(self):
        self.engine = None  # type: LTorrentEngine
        self.status = 'ok'
        self._api = {
            'status': self.api_status,
            'open_torrent': self.api_open_torrent,
            'open': self.api_open,
            'get_file_list': self.api_get_file_list,
        }

    def bind_engine(self, engine):
        self.engine = engine  # type: LTorrentEngine

    def api(self, req):
        if 'action' not in req:
            return {'result': False}
        if req['action'] in self._api:
            return self._api[req['action']](req)
        return {'result': False}

    def api_open(self, req):
        if 'f_index' not in req:
            return {'result': False}
        self.engine.f_open(int(req['f_index']))
        return {'result': True}

    def api_open_torrent(self, req):
        if 'url' in req:
            logger.info('Open torrent from url %s' % req['url'])
            raw_data = HTTPClient().get_page(req['url'])
        elif 'file' in req:
            logger.info('Open torrent from file %s' % req['file'])
            raw_data = open(req['file']).read()
        elif 'raw_data' in req:
            logger.info('Open torrent from raw data')
            raw_data = req['raw_data']
        else:
            raise Exception('Error in api_open_file %s' % req)
        info_hash = self.engine.open_torrent(raw_data)
        # web_server.engine
        return {'result': True, 'info_hash': info_hash}

    def api_get_file_list(self, req):
        return {'result': True, 'file_list': self.engine.cache.torrent.get_files()}

    def api_status(self, req):
        return {'result': True, 'status': self.status}


web_server = WebServer()


class HMain(tornado.web.RequestHandler):
    def get(self):
        now = time.time()
        self.render('index.html', engine=web_server.engine, now=now)


class HStatus(tornado.web.RequestHandler):
    def get(self):
        now = time.time()
        if web_server.engine.cache is None:
            self.write('Open torrent first')
            return
        self.render('status.html', engine=web_server.engine, now=now)


class HUploadTorrent(tornado.web.RequestHandler):
    def post(self):
        try:
            raw_data = self.request.files['torrent'][0]['body']
            info_hash = web_server.engine.open_torrent(raw_data)
            self.write(json.dumps({'result': True, 'info_hash': info_hash}))
        except Exception, e:
            logger.exception('upload torrent error')
            self.write(json.dumps({'result': False}))


class HAPI(tornado.web.RequestHandler):
    def get(self):
        self.write('API')

    def post(self):
        self.set_header('Content-Type', 'application/json')
        self.write(json.dumps(web_server.api(tornado.escape.json_decode(self.request.body))))


class HPlay(tornado.web.RequestHandler):
    def head(self, info_hash, f_index, f_name):
        if web_server.engine.cache is None or web_server.engine.cache.torrent.info_hash().encode('hex') != info_hash:
            self.send_error(404)
            return
        self.set_header('Content-Length', web_server.engine.f_size)
        self.set_header('Accept-Ranges', 'bytes')
        self.set_status(200)


    @tornado.gen.coroutine
    def get(self, info_hash, f_index, f_name):
        if web_server.engine.cache is None or web_server.engine.cache.torrent.info_hash().encode('hex') != info_hash:
            self.send_error(404)
            return
        f_index = int(f_index)
        web_server.engine.f_open(f_index)
        # self.set_header('Content-Type', 'video/mkv')
        self.set_header('Content-Disposition', 'attachment; filename=%s' % web_server.engine.cache.torrent.get_files()[f_index])
        get_range = self.request.headers.get('Range', None)
        if get_range is not None:
            logger.info('WEB_SERVER range %s' % get_range)
            get_range = re_parce_range.findall(get_range)[0]
            logger.info('WEB_SERVER range %s, header %s' % (str(get_range), str(self.request.headers.get('Range', None))))
        file_size = web_server.engine.f_size
        if get_range is not None:
            offset = int(get_range[0])
            web_server.engine.f_seek(offset)
            self.set_status(206)
            self.add_header('Content-Range', 'bytes %s-%s/%s' % (offset, file_size, file_size))
            self.set_header('Content-Length', file_size - offset)
        else:
            self.set_header('Content-Length', file_size)
        # send data
        while True:
            try:
                buf = web_server.engine.f_read(WEB_SERVER_SEND_BUF)
            except EWaitPiece as e:
                if e.wait_piece in web_server.engine.d_piece_conditions:
                    piece_condition = web_server.engine.d_piece_conditions[e.wait_piece]
                else:
                    piece_condition = Condition()
                    web_server.engine.d_piece_conditions[e.wait_piece] = piece_condition
                yield piece_condition.wait()
                buf = web_server.engine.f_read(WEB_SERVER_SEND_BUF)
            if len(buf) == 0:
                break
            self.write(buf)
            try:
                yield self.flush()
            except StreamClosedError:
                logger.info('WEB_SERVER client close connection')
                return


def make_web_app():
    app = tornado.web.Application([
        (r'/', HMain),
        (r'/api', HAPI),
        (r'/status', HStatus),
        (r'/play/(\w+)/(\d+)/(.*)', HPlay),
        (r'/upload_torrent/', HUploadTorrent),
        (r'/static/(.*)', tornado.web.StaticFileHandler, {'path': os.path.join(MY_DIR, 'static')}),
    ])
    app.settings['debug'] = True
    app.settings['compiled_template_cache'] = False
    app.settings['template_path'] = os.path.join(MY_DIR, 'templates')
    logger.info('Start web server on port %s' % WEB_SERVER_PORT)
    app.listen(WEB_SERVER_PORT)
    return app
